/* psoc-ctsic.c
 *
 * Control driver for PSOC Capacitive TouchScreen IC
 *
 * Copyright (C) 2006 TomTom BV <http://www.tomtom.com/>
 * Authors: Vincent Dejouy <vincent.dejouy@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/cdev.h>
#include <linux/i2c.h>
#include <linux/workqueue.h>
#include <linux/spinlock.h>
#include <linux/err.h>
#include <linux/psoc-ctsic.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/input.h>
#include <linux/timer.h>
#include <linux/jiffies.h>
#include <linux/firmware.h>
#include <linux/kernel.h>
#include <linux/moduleparam.h>
#include <linux/jiffies.h>
#include <plat/fdt.h>

#define VERBOSE_LEVEL    VERBOSE_ERROR

static struct _psoc_i2c {
	struct i2c_client	*client;
	struct work_struct	work;
} psoc_bootloader_i2c;


static struct firmware psoc_firmware;
static int index_firmware=0;

static char current_state = STATE_INITIAL;

static char buff_temp_ascii[MAX_CHAR_PER_LINE];
static char buff_temp_byte[MAX_BYTE_PER_LINE];


static char bootloader_key_fw_v1[] = {0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07};
static char bootloader_key_fw_v2[] = {0x4E, 0x69, 0x4C, 0x61, 0x56, 0x69, 0x44, 0x65};

static char* action="none";
static char* rst="false";

/**
 * psoc_bootloader_i2c_recv - issue a single I2C message in master receive mode
 * @buf: Where to store data read from slave
 * @count: How many bytes to read
 *
 * Returns negative errno, or else the number of bytes read.
 */
static int psoc_bootloader_i2c_recv(char *buf, int count)
{
	int ret = -ENODEV;
	BUG_ON(!buf);
	ret = i2c_master_recv(psoc_bootloader_i2c.client, buf, count);
	if ( ret != count )
		printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": Error reading bootloader registers, return value of i2c read: [%d] instead of [%d]\n", ret, count);
	return ret;
}


/**
 * psoc_bootloader_i2c_send - issue a single I2C message in master receive mode
 * @buf: Message to be written to the slave
 * @count: How many bytes to read
 *
 * Returns negative errno, or else the number of bytes read.
 */
static int psoc_bootloader_i2c_send(char *buf, int count, int verbose)
{
	int ret = -ENODEV;
	int i;

	BUG_ON(!buf);

	ret = i2c_master_send(psoc_bootloader_i2c.client, buf, count);
	if ( ret != count )
		printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": Error writing bootloader registers, return value of i2c send: [%d] instead of [%d]\n", ret,count);

	mdelay(10);

	if (verbose == VERBOSE_DEBUG) {
		printk("buff send to bootloader of cypress cap touch screen:\n");
		for (i=0;i<count;i++) {
			printk("0x%x ", buf[i]);
		}
		printk("\n");
		printk("Result of send i2c cmd: %d\n", ret);
	}

	return ret;
}

//Returns 0 if erase not finished
//Returns 1 if erase finished
static int psoc_bootloader_flash_erase_finished(int verbose)
{
	int ret;
	char reg_addr = STATUS_REGISTER;
	char buff[3];
	int i;

	ret = psoc_bootloader_i2c_send(&reg_addr, sizeof(char), verbose);
	if ( ret != sizeof(char) ){
		printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": Error writing 1 byte to the touch screen\n");
		return -1;
	}

	udelay(200);
	
	ret = psoc_bootloader_i2c_recv(buff, 3*sizeof(char));
	if ( ret != 3*sizeof(char) ){
		printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": Error reading 3 bytes from the touch screen\n");
		return -1;
	}

	if (verbose == VERBOSE_DEBUG){
		printk("Result from getting flash erase status:\n");
		for (i=0;i<3;i++)
		{
			printk("0x%x ", buff[i]);
		}

		printk("\nFlash erase status Read = 0x%X\n", buff[1]);
	
	}

	if ( ( buff[1] & 0x80 ) != 0 ){
		return 0;
	} else {
		return 1;
	}
}


static int psoc_bootloader_read_status(int verbose)
{
	int ret;
	char reg_addr = STATUS_REGISTER;
	char buff[3];
	int i;

	ret = psoc_bootloader_i2c_send(&reg_addr, sizeof(char), verbose);
	if ( ret != sizeof(char) ){
		printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ":Error writing 1 byte to the touch screen\n");
		return -1;
	}

	udelay(200);
	
	ret = psoc_bootloader_i2c_recv(buff, 3*sizeof(char));
	if ( ret != 3*sizeof(char) ){
		printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ":Error reading 3 bytes from the touch screen\n");
		return -1;
	}

	if (verbose == VERBOSE_DEBUG){
		printk("Result from getting status:\n");
		for (i=0;i<3;i++)
		{
			printk("0x%x ", buff[i]);
		}

		printk("\nStatus Read = 0x%X\n", buff[2]);
	}

	if ( buff[2] == 0x20 ){
		return 0;
	} else {
		printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": Status incorrect, Result from getting status:\n");
		for (i=0;i<3;i++)
			printk("0x%x ", buff[i]);

		printk("\nStatus Read = 0x%X\n", buff[2]);
		return (int) buff[2];
	}

}

static bool psoc_bootloader_isadigit(char c)
{
	if ( ( c >= '0' ) && ( c <= '9' ) )
		return true;
	else
		return false;
}

static char psoc_bootloader_atoc(char* buff)
{
	char res;

	if ( psoc_bootloader_isadigit(*(buff+1)) == true )
		res = *(buff+1) - '0';
	else
		res = *(buff+1) - 'A' + 10;

	if ( psoc_bootloader_isadigit(*(buff)) == true )
		res += (*(buff) - '0')*16;
	else
		res += (*(buff) - 'A' + 10)*16;

	return res;

}

static int psoc_bootloader_extract_line(char* buffer, char* last_line, int verbose)
{
	int i=0;

	while( (*psoc_firmware.data != '\n') && (*psoc_firmware.data != '\r')){
		if (index_firmware == (int) psoc_firmware.size) {
			if (verbose == VERBOSE_DEBUG)
				printk("end of file detected\n");
			*last_line = 1;
			break;
		}
		buffer[i++]=*(psoc_firmware.data++);
		index_firmware++;
	}

	buffer[i]=0;

	if (!i){
		psoc_firmware.data++;
		index_firmware++;
	}

	return i;
}

static void psoc_bootloader_convert_short_to_buff(short number_to_convert, char** output_buffer)
{
	(*output_buffer)[0]=(char)(number_to_convert >> 8);
	(*output_buffer)[1]=(char)(number_to_convert);
}

static int psoc_bootloader_state_generic(char* buff, int length, int verbose)
{

	int i,j;
	char buff_to_send[MAX_LENGTH_I2C_FRAME + SIZE_PREFIX_BOOTLOADER_CMD];
	int length_last_line;
	int length_line;
	int quantity_of_lines;
	short prefix=0;
	char buff_prefix[SIZE_PREFIX_BOOTLOADER_CMD];
	int addr_buff_prefix = (int) buff_prefix;

	length_last_line = length%MAX_LENGTH_I2C_FRAME;
	quantity_of_lines = (length/MAX_LENGTH_I2C_FRAME) + (length_last_line ? 1: 0);

	for (i=0;i<quantity_of_lines;i++)
	{
		psoc_bootloader_convert_short_to_buff(prefix, (char**) &addr_buff_prefix);
		buff_to_send[0]=buff_prefix[0];
		buff_to_send[1]=buff_prefix[1];
		
		if ( i != (quantity_of_lines-1))
			length_line=MAX_LENGTH_I2C_FRAME;
		else
			length_line=length_last_line;

		for (j=0;j<length_line;j++)
			buff_to_send[j+SIZE_PREFIX_BOOTLOADER_CMD] = buff[i*MAX_LENGTH_I2C_FRAME + j];

		psoc_bootloader_i2c_send(buff_to_send, length_line+SIZE_PREFIX_BOOTLOADER_CMD, verbose);
		prefix+=0x10;
	}

	return 0;

}

static int psoc_bootloader_state_write_block(char* buff, int length, int verbose)
{
	int ret;

	if (verbose == VERBOSE_DEBUG){
		printk("Write Block %.2X%.2X\n", buff[10], buff[11]);
	}
	psoc_bootloader_state_generic(buff,length,verbose);
	msleep(100);

	ret = psoc_bootloader_read_status(verbose);
	if ( ret )
		printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": Error of status in state write block\n");

	return ret;
}

static int psoc_bootloader_state_enter_bootloader(char* buff, int length, int verbose)
{
	int ret;
	int retries;
	
	if (verbose == VERBOSE_DEBUG){
		printk("Enter Bootloader Mode\n");
		printk("Key: %.2X%.2X%.2X%.2X%.2X%.2X%.2X%.2X\n", buff[2], buff[3], buff[4], buff[5], buff[6], buff[7], buff[8], buff[9]);
	}
	psoc_bootloader_state_generic(buff,length,verbose);
	msleep(300);

	retries=20;
	while ( !psoc_bootloader_flash_erase_finished(VERBOSE_DISABLED) )
	{
		msleep(300);
		retries--;
		if ( !retries )
			break;
	}

	ret = psoc_bootloader_read_status(verbose);
	if ( ret )
		printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": Error of status in state enter bootloader\n");

	return ret;
}

static int psoc_bootloader_state_exit_bootloader(char* buff, int length, int verbose)
{
	if (verbose == VERBOSE_DEBUG) {
		printk("Exit Bootloader Mode\n");
		printk("Key: %.2X%.2X%.2X%.2X%.2X%.2X%.2X%.2X\n", buff[2], buff[3], buff[4], buff[5], buff[6], buff[7], buff[8], buff[9]);
	}
	psoc_bootloader_state_generic(buff,length,verbose);

	return 0;
}

static int psoc_bootloader_execute_state(char state, char* buffer, int length, int verbose)
{
	int ret=0;

	switch (state) {
		case STATE_ENTER_BOOTLOADER: ret=psoc_bootloader_state_enter_bootloader(buffer, length, verbose);
					break;
		case STATE_WRITE_BLOCK: ret=psoc_bootloader_state_write_block(buffer, length, verbose);
					break;
		case STATE_EXIT_BOOTLOADER: ret=psoc_bootloader_state_exit_bootloader(buffer, length, verbose);
					break;
		default: ret=ERROR_UNKNOWN_STATE;
	}

	return ret;
}


static int psoc_bootloader_process_line(char* buffer, int length, int verbose)
{
	char cmd;

	if ( buffer[0] != BOOTLOADER_CMD_PREFIX )
		return ERROR_INVALID_FIRMWARE_FORMAT;

	cmd = buffer[1];

	switch (cmd) {
		case ENTER_BOOTLOADER_CMD:
		case WRITE_BLOCK_CMD:
		case EXIT_BOOTLOADER_CMD: break;
		default: return ERROR_UNKNOWN_CMD;
	}

	switch (current_state) {
		case STATE_INITIAL: 		if ( cmd == ENTER_BOOTLOADER_CMD )
							current_state = STATE_ENTER_BOOTLOADER;
						else
							return ERROR_FIRMWARE_DOES_NOT_START_WITH_ENTER_BL_CMD;
						break;
		case STATE_ENTER_BOOTLOADER: 	if ( cmd == ENTER_BOOTLOADER_CMD )
							current_state = STATE_ENTER_BOOTLOADER;
						else if ( cmd == WRITE_BLOCK_CMD )
							current_state = STATE_WRITE_BLOCK;
						else
							return ERROR_INVALID_CMD_SEQUENCE;
						break;
		case STATE_WRITE_BLOCK: 	if ( cmd == WRITE_BLOCK_CMD )
							current_state = STATE_WRITE_BLOCK;
						else if ( cmd == EXIT_BOOTLOADER_CMD )
							current_state = STATE_EXIT_BOOTLOADER;
						else
							return ERROR_INVALID_CMD_SEQUENCE;
						break;
		case STATE_EXIT_BOOTLOADER: 	if ( cmd == EXIT_BOOTLOADER_CMD )
							current_state = STATE_EXIT_BOOTLOADER;
						else
							return ERROR_INVALID_CMD_SEQUENCE;
						break;
		default:			return ERROR_UNKNOWN_STATE;
	}

	if (verbose == VERBOSE_DEBUG)
		printk("\nExecute state %d\n", current_state);
	return psoc_bootloader_execute_state(current_state, buffer, length, verbose);

}

static int psoc_bootloader_parse_lines(int verbose)
{
	int nr_chars_line = 0;
	int nr_bytes_line=0;
	char flag_last_line=0;
	int i,j,k;
	int ret;

	k=0;

	while (!flag_last_line){

		nr_chars_line = psoc_bootloader_extract_line(buff_temp_ascii, &flag_last_line, verbose);

		if (nr_chars_line) {

			nr_bytes_line=nr_chars_line/2;
			j=0;
			for (i=0;i<nr_chars_line;i+=2)
				buff_temp_byte[j++] = psoc_bootloader_atoc( buff_temp_ascii + i );

			ret = psoc_bootloader_process_line(buff_temp_byte, nr_bytes_line, verbose);

			if (ret){
				printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": Error processing line %d\n", k);
				printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": Status Returned by the touch screen: 0x%X, instead of 0x20\n", ret);
				return -1;
			}

			k++;

		}
	}

	return 0;

}


static void psoc_bootloader_loading_finished(const struct firmware *fw, void* context)
{

	int ret;

	printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": request firmware finished\n");
	if ( fw != NULL )
		memcpy((void*) &psoc_firmware, (void*) fw, sizeof(struct firmware));
	else
		return;

	printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": firmware received\n");
	printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": update in progress...\n");

	ret = psoc_bootloader_parse_lines(VERBOSE_LEVEL);
	if (ret == 0 )
		printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": update of the capacitive touch screen firmware successful ...\n");
	else
		printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": update of the capacitive touch screen firmware failed ...\n");

}

static void psoc_bootloader_work(struct work_struct *work)
{
	int ret;
	ret = request_firmware_nowait(THIS_MODULE, FW_ACTION_HOTPLUG, "fw_cap_ts.dld", &psoc_bootloader_i2c.client->dev, NULL, psoc_bootloader_loading_finished);
	if (ret != 0)
		printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": Request firmware failed, return code = %d\n", ret);
}

static void psoc_bootloader_i2c_init(struct i2c_client *client)
{
	BUG_ON(!client);
	//Init psoc_i2c struct
	memset(&psoc_bootloader_i2c, 0x00, sizeof(psoc_bootloader_i2c));
	psoc_bootloader_i2c.client = client;
	INIT_WORK(&psoc_bootloader_i2c.work, psoc_bootloader_work);
}

static int psoc_bootloader_i2c_start_ts(void)
{

	char *buff_temp;
	int size_buff_temp;
	int ret;
	char* ts = NULL;

	printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": Start the touch screen application\n");
	size_buff_temp = SIZE_PREFIX_BOOTLOADER_CMD + SIZE_BOOTLOADER_CMD + SIZE_BOOTLOADER_KEY;
	buff_temp = (char*) kmalloc(size_buff_temp*sizeof(char*), GFP_KERNEL);

	//Command prefix
	buff_temp[0]=0x00;
	buff_temp[1]=0x00;

	//Bootloader command
	buff_temp[2]=BOOTLOADER_CMD_PREFIX;
	buff_temp[3]=RUN_APP_CMD;


	ts = fdt_get_string ("/features", "ts", "");

	//Bootloader key
	if ( strstr(ts, "FW_V2") != NULL )
		memcpy((void*) (buff_temp + 4), (void*) bootloader_key_fw_v2, SIZE_BOOTLOADER_KEY*sizeof(char));
	else
		memcpy((void*) (buff_temp + 4), (void*) bootloader_key_fw_v1, SIZE_BOOTLOADER_KEY*sizeof(char));

	ret=psoc_bootloader_i2c_send(buff_temp, size_buff_temp, VERBOSE_DEBUG);
	if ( ret != size_buff_temp )
		return -1;
	kfree(buff_temp);
	return 0;
}

static int psoc_bootloader_i2c_probe(struct i2c_client *client, const struct i2c_device_id *id)
{

	psoc_pdata_t    *pdata;

	printk(KERN_INFO PSOC_BOOTLOADER_DEVNAME ": probing driver\n");

	index_firmware=0;

	pdata = (psoc_pdata_t *)client->dev.platform_data;

	printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": Psoc Bootloader Capacitive Touch Screen Driver, (C) 2008 TomTom BV\n" );

	BUG_ON(!client);
	psoc_bootloader_i2c_init(client);

	if (strncmp(rst, "true", 4) == 0) {
		pdata->reset_ts();
		msleep(DELAY_BETWEEN_RELEASE_RST_AND_SEND_1ST_CMD);
		printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": Capacitive Touch Screen Reset\n" );
	}

	if (strncmp(action, "start_ts", 8) == 0) {
		return (psoc_bootloader_i2c_start_ts());
	} else if (strncmp(action, "start_update", 12) == 0) {
		printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": Start the update firmware application\n");
		schedule_work(&psoc_bootloader_i2c.work);
	} else {
		printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": Error, the argument \"action\" is not set properly\n");
		printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": It should be either:\n");
		printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": - \"start_ts\"\n");
		printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": or\n");
		printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": - \"start_update\"\n");
		printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": Remove the driver and re modprobe it with the correct arguments\n");
		return -1;
	}

	return 0;
}

static int psoc_bootloader_i2c_remove(struct i2c_client *client)
{
	printk(KERN_INFO PSOC_BOOTLOADER_DEVNAME ": start remove\n");
	return 0;
}

static int psoc_bootloader_i2c_suspend(struct i2c_client *client, pm_message_t mesg)
{
	printk(KERN_INFO PSOC_BOOTLOADER_DEVNAME ": start suspending\n");
	return 0;
}

static int psoc_bootloader_i2c_resume(struct i2c_client *client)
{
	printk(KERN_INFO PSOC_BOOTLOADER_DEVNAME ": start resuming\n");
	return (psoc_bootloader_i2c_start_ts());
}

static const struct i2c_device_id psoc_bootloader_id[] = {
	{ PSOC_BOOTLOADER_DEVNAME, 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, psoc_bootloader_id);

static struct i2c_driver psoc_bootloader_i2c_driver=
{
	.probe	= psoc_bootloader_i2c_probe,
	.remove	= psoc_bootloader_i2c_remove,
	.suspend= psoc_bootloader_i2c_suspend,
	.resume	= psoc_bootloader_i2c_resume,
	.id_table = psoc_bootloader_id,
	.driver = {
		.name	= PSOC_BOOTLOADER_DEVNAME,
		.owner	= THIS_MODULE,
	},
};

static int __init psoc_bootloader_init(void)
{
	int err;

	printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": start init\n");

	if ((err = i2c_add_driver(&psoc_bootloader_i2c_driver))) {
		printk(KERN_ERR PSOC_BOOTLOADER_DEVNAME ": Could Not Be Added. Err Code: [%i]\n", err);
	}

	return err;
}

static void __exit psoc_bootloader_exit(void)
{
	printk(KERN_INFO PSOC_BOOTLOADER_DEVNAME ": start exit\n");
	i2c_del_driver(&psoc_bootloader_i2c_driver);
}

MODULE_AUTHOR("Vincent Dejouy <vincent.dejouy@tomtom.com>");
MODULE_DESCRIPTION("Driver for Bootloader Of PSoc, Capacitive Touch Screen Controller.");
MODULE_LICENSE("GPL");

module_init(psoc_bootloader_init);
module_exit(psoc_bootloader_exit);
module_param(action, charp, S_IRUGO);
module_param(rst, charp, S_IRUGO);

