/* psoc-ctsic.c
 *
 * Control driver for PSOC Capacitive TouchScreen IC
 *
 * Copyright (C) 2006 TomTom BV <http://www.tomtom.com/>
 * Authors: Vincent Dejouy <vincent.dejouy@tomtom.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/cdev.h>
#include <linux/i2c.h>
#include <linux/workqueue.h>
#include <linux/spinlock.h>
#include <linux/err.h>
#include <linux/psoc-ctsic.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/input.h>
#include <linux/timer.h>
#include <linux/jiffies.h>


#define VERBOSE_LEVEL					VERBOSE_ERROR

static struct _psoc_i2c {
        struct i2c_client       *client;
        struct work_struct      read_info_work;
	struct work_struct	polling_work;
} psoc_i2c;

I2C_DATA_RECEIVED_STRUCT i2c_packet_received;
I2C_DATA_RECEIVED_SYS_INFO_STRUCT i2c_packet_received_sys_info;
I2C_DATA_RECEIVED_RAW_COUNTS_STRUCT i2c_packet_received_raw_counts;

static int current_mode = OPERATING_MODE;
static struct timer_list mytimer;
static int len_i2c_frame_finger_position;
static int len_i2c_frame_sys_info;
static int len_i2c_frame_test_mode;
static char header_data_tunning[]={0x55, 0xAA};
static char footer_data_tunning[]={0xFF, 0xFF};
static psoc_pdata_t pdata;
static struct input_dev *psoc_cap_ts_input_dev;     /* Representation of an input device */
static psoc_drv_data_t *drv_data;
static char* recalibrate="false";

static inline unsigned short psoc_swap_lsb_msb(char* number_to_convert)
{
	return ( (number_to_convert[1]) | ( number_to_convert[0] << 8) );
}

/**
 * psoc_i2c_send - issue a single I2C message in master transmit mode
 * @buf: Data that will be written to the slave
 * @count: How many bytes to write
 *
 * Returns negative errno, or else the number of bytes written.
 */
static int psoc_i2c_send(const char *buf, int count)
{
	int ret = -ENODEV;

	BUG_ON(!buf);
	
	ret = i2c_master_send(psoc_i2c.client, buf, count);

	return ret;
}


/**
 * psoc_i2c_recv - issue a single I2C message in master receive mode
 * @buf: Where to store data read from slave
 * @count: How many bytes to read
 *
 * Returns negative errno, or else the number of bytes read.
 */
static int psoc_i2c_recv(char *buf, int count)
{
	int ret = -ENODEV;

	BUG_ON(!buf);

	ret = i2c_master_recv(psoc_i2c.client, buf, count);
	
	return ret;
}

//These are a lot of hard coded values but it's not necessary to make this function beautiful because
//it's only required to use it for intermediary test and old firmware
static void psoc_re_calibrate(void)
{
	char buff_temp[3];
        int ret;

	msleep(2000);

	//Switch to sys info mode
	buff_temp[0]=0x00;
	buff_temp[1]=0x10;

	ret=psoc_i2c_send(buff_temp, 2);
        if ( ret != 2 )
                return;

	msleep(1000);

	//Run mfg test command
	buff_temp[0]=0x00;
	buff_temp[1]=0x01;
	buff_temp[2]=0x30;

	ret=psoc_i2c_send(buff_temp, 3);
        if ( ret != 3 )
                return;

	msleep(1000);
}

static int psoc_start_ts_app(void)
{

	char *buff_temp;
	int size_buff_temp;
	int ret;

	printk(KERN_INFO PSOC_DEVNAME ": Start the touch screen application\n");
	size_buff_temp = SIZE_PREFIX_BOOTLOADER_CMD + SIZE_BOOTLOADER_CMD + SIZE_BOOTLOADER_KEY;
	buff_temp = (char*) kmalloc(size_buff_temp*sizeof(char), GFP_KERNEL);

	//Command prefix
	buff_temp[0]=0x00;
	buff_temp[1]=0x00;

	//Bootloader command
	buff_temp[2]=BOOTLOADER_CMD_PREFIX;
	buff_temp[3]=RUN_APP_CMD;

	//Bootloader key
	memcpy((void*) (buff_temp + 4), (void*) pdata.key, SIZE_BOOTLOADER_KEY*sizeof(char));

	ret=psoc_i2c_send(buff_temp, size_buff_temp);
	if ( ret != size_buff_temp )
	{
		kfree(buff_temp);
		return -1;
	}
	
	current_mode = OPERATING_MODE;

	kfree(buff_temp);
	return 0;
}

static int psoc_restart_ts(void)
{
	pdata.reset_ts();
	msleep(DELAY_BETWEEN_RELEASE_RST_AND_SEND_1ST_CMD);
	return (psoc_start_ts_app());
}

static int psoc_get_finger_position(int verbose)
{

	int ret;
	int i;
	char touches;
	char* buff_temp;
	unsigned short temp;

	buff_temp = (char*) kmalloc(len_i2c_frame_finger_position, GFP_KERNEL);
	if (buff_temp == NULL){
		if (verbose)
			printk(KERN_ERR PSOC_DEVNAME ": Error reading finger positions, can't allocate a buffer of [%d] bytes\n", len_i2c_frame_finger_position);
		return 1;
	}

	ret = psoc_i2c_recv(buff_temp, len_i2c_frame_finger_position);

	if (ret != len_i2c_frame_finger_position){
		if (verbose)
			printk(KERN_ERR PSOC_DEVNAME ": Error reading finger positions, return value of i2c read: [%d] instead of [%d]\n", ret, len_i2c_frame_finger_position);
		kfree(buff_temp);
		return ERROR_I2C_ACCESS;
	}

	if ( (verbose == VERBOSE_DEBUG) || (verbose == VERBOSE_TUNING) ){
		printk("\n");
		printk(PSOC_DEVNAME ":Get finger position, raw data received:\n");
		printk("%.2x %.2x %.2x ", header_data_tunning[0], header_data_tunning[1], len_i2c_frame_finger_position);
		for (i=0;i<len_i2c_frame_finger_position;i++){
			printk("%.2x ",buff_temp[i]);
		}
		printk("%.2x %.2x\n", footer_data_tunning[0], footer_data_tunning[1]);
		printk("\n");
	}

	memcpy((void*) &i2c_packet_received, (void*) buff_temp, len_i2c_frame_finger_position);

	temp = psoc_swap_lsb_msb((char*) &i2c_packet_received.finger1_position.x);
	i2c_packet_received.finger1_position.x = temp;
	temp = psoc_swap_lsb_msb((char*) &i2c_packet_received.finger1_position.y);
	i2c_packet_received.finger1_position.y = temp;

	temp = psoc_swap_lsb_msb((char*) &i2c_packet_received.finger2_position.x);
	i2c_packet_received.finger2_position.x = temp;
	temp = psoc_swap_lsb_msb((char*) &i2c_packet_received.finger2_position.y);
	i2c_packet_received.finger2_position.y = temp;

	if (verbose == VERBOSE_DEBUG){

		printk(PSOC_DEVNAME ":i2c packet received after finger positions reversed:\n");
		printk("HST_MODE: 0x%x\n", i2c_packet_received.hst_mode);
		printk("TT_MODE: 0x%x\n", i2c_packet_received.tt_mode);

		printk("\nFinger 1:\n");\
		printk("X: 0x%x\n", i2c_packet_received.finger1_position.x);
		printk("Y: 0x%x\n", i2c_packet_received.finger1_position.y);
		printk("Z: 0x%x\n", i2c_packet_received.finger1_position.z);

		printk("\nFinger 2:\n");
		printk("X: 0x%x\n", i2c_packet_received.finger2_position.x);
		printk("Y: 0x%x\n", i2c_packet_received.finger2_position.y);
		printk("Z: 0x%x\n", i2c_packet_received.finger2_position.z);

		printk("\nFingers: %d\n", i2c_packet_received.large_objects_detect_and_touches & TT_STAT_REGISTER_TOUCHES_MASK);
		printk("Gesture Count: %d\n", i2c_packet_received.gesture_count_and_overflow_flag & GEST_COUNT_REGISTER_GEST_COUNT_MASK);
		printk("Gesture ID: 0x%x\n", i2c_packet_received.gestureID);
		printk("\n");

	}


	touches = i2c_packet_received.large_objects_detect_and_touches & TT_STAT_REGISTER_TOUCHES_MASK;

	//Sometimes the TS reports more than max number of touches, they should not be reported to the input layer
	//Because in the datasheet it is mentioned that number of touches greater than max touches are "reserved"
	//Most probably debug purposes for Cypress
	if (touches <= MAXNUMTOUCHES){
		//Report position of fingers to the input layer
		input_report_abs(psoc_cap_ts_input_dev, ABS_X, i2c_packet_received.finger1_position.x);
		input_report_abs(psoc_cap_ts_input_dev, ABS_Y, i2c_packet_received.finger1_position.y);
		input_report_abs(psoc_cap_ts_input_dev, ABS_FINGER2_X, i2c_packet_received.finger2_position.x);
		input_report_abs(psoc_cap_ts_input_dev, ABS_FINGER2_Y, i2c_packet_received.finger2_position.y);

		input_report_abs(psoc_cap_ts_input_dev, ABS_GESTURE, i2c_packet_received.gestureID);

		input_report_abs(psoc_cap_ts_input_dev, ABS_FINGERS, i2c_packet_received.large_objects_detect_and_touches & TT_STAT_REGISTER_TOUCHES_MASK);

		if (touches)
			input_report_abs(psoc_cap_ts_input_dev, ABS_PRESSURE, 1);
		else
			input_report_abs(psoc_cap_ts_input_dev, ABS_PRESSURE, 0);

		input_sync(psoc_cap_ts_input_dev);
	}

	kfree(buff_temp);

	return 0;

}

static int psoc_set_offset_register(int verbose, char offset)
{
	int ret;

	ret = psoc_i2c_send(&offset, 1);

	if (verbose == VERBOSE_DEBUG){
		printk(PSOC_DEVNAME ":set offset register to:\n");
		printk("0x%x ",offset);
		printk("\n");
	}

	if (ret != 1 && verbose)
		printk(KERN_ERR PSOC_DEVNAME ": Error setting offset register, return value of i2c send: [%d] instead of 1\n", ret);
          
	return (ret == 1)?0:ERROR_I2C_ACCESS;

}

static int psoc_read_register(int verbose, char* value)
{
	int ret;

	//A register is 1 byte so we read 1 byte
	ret = psoc_i2c_recv(value, 1);

	if (verbose == VERBOSE_DEBUG){
		printk(PSOC_DEVNAME ":register value read:\n");
		printk("0x%x ",*value);
		printk("\n");
	}

	if (ret != 1 && verbose)
		printk(KERN_ERR PSOC_DEVNAME ": Error reading register value, return value of i2c read: [%d] instead of 1\n", ret);

	return (ret == 1)?0:ERROR_I2C_ACCESS;

}

static int psoc_write_register(int verbose, char reg, char value)
{
	int ret;
	char buff[2];

	buff[0]=reg;
	buff[1]=value;

	ret = psoc_i2c_send(buff, 2);

	if (verbose == VERBOSE_DEBUG){
		printk(PSOC_DEVNAME ":register offset = 0x%X value set: 0x%X\n", reg, value);
	}

	if (ret != 2 && verbose)
		printk(KERN_ERR PSOC_DEVNAME ": Error setting register value, return value of i2c send: [%d] instead of 2\n", ret);

	return (ret == 2)?0:ERROR_I2C_ACCESS;

}


static int psoc_switch_to_other_mode(int verbose, char device_mode)
{
	char hst_mode_register_offset=HST_MODE_REGISTER_OFFSET;
	char hst_mode_register_value;
	char new_hst_mode_register_value=0;

	if (verbose == VERBOSE_DEBUG)
		printk("Enter in switch to other mode, target mode=%d\n", device_mode);

	//Set offset of HOST MODE register
	if (psoc_set_offset_register(verbose, hst_mode_register_offset))
		return ERROR_I2C_ACCESS;

	udelay(DELAY_BETWEEN_SET_OFFSET_REG_AND_RW_REG_VALUE);

	//Read value of HOST MODE register
	if (psoc_read_register(verbose, &hst_mode_register_value))
		return ERROR_I2C_ACCESS;

	new_hst_mode_register_value = (hst_mode_register_value & ~HST_MODE_REGISTER_DEVICE_MODE_MASK ) | (device_mode << HST_MODE_REGISTER_DEVICE_MODE_OFFSET);

	if (verbose == VERBOSE_DEBUG){	
		printk("Switch to other mode, Old hst mode register value: 0x%x\n", hst_mode_register_value);
		printk("Switch to other mode, New hst mode register value: 0x%x\n", new_hst_mode_register_value);
	}

	//Write value of HOST MODE register
	if (psoc_write_register(verbose, hst_mode_register_offset, new_hst_mode_register_value))
		return ERROR_I2C_ACCESS;

	current_mode = (int) device_mode;

	if (verbose == VERBOSE_DEBUG)
		printk("Exit switch to other mode\n");

	return 0;

}


static int psoc_toggle_data_read_bit(int verbose)
{
	char hst_mode_register_offset=HST_MODE_REGISTER_OFFSET;
	char hst_mode_register_value;
	char new_hst_mode_register_value=0;
	char data_toggle_read_bit=0;

	if (verbose == VERBOSE_DEBUG)
		printk("Enter in toggle data read bit\n");

	//Set offset of HOST MODE register
	if (psoc_set_offset_register(verbose, hst_mode_register_offset))
		return ERROR_I2C_ACCESS;

	udelay(DELAY_BETWEEN_SET_OFFSET_REG_AND_RW_REG_VALUE);

	//Read value of HOST MODE register
	if (psoc_read_register(verbose, &hst_mode_register_value))
		return ERROR_I2C_ACCESS;

	data_toggle_read_bit = hst_mode_register_value >> HST_MODE_REGISTER_DATA_RD_TOGGLE_OFFSET;

	new_hst_mode_register_value = hst_mode_register_value;

	new_hst_mode_register_value &= ~HST_MODE_REGISTER_DATA_RD_TOGGLE_MASK;
	new_hst_mode_register_value |= ~data_toggle_read_bit << HST_MODE_REGISTER_DATA_RD_TOGGLE_OFFSET; 

	if (verbose == VERBOSE_DEBUG){	
		printk("Toggle data read bit, Old hst mode register value: 0x%x\n", hst_mode_register_value);
		printk("Toggle data read bit, New hst mode register value: 0x%x\n", new_hst_mode_register_value);
	}

	//Write value of HOST MODE register
	if (psoc_write_register(verbose, hst_mode_register_offset, new_hst_mode_register_value))
		return ERROR_I2C_ACCESS;

	if (verbose == VERBOSE_DEBUG)
		printk("Exit from toggle data read bit\n");

	return 0;
}


static int psoc_read_mode(int verbose)
{
	char tt_mode_register_offset=TT_MODE_REGISTER_OFFSET;
	char tt_mode_register_value;

	if (verbose == VERBOSE_DEBUG)
		printk("Enter in read mode\n");

	//Set offset of TT MODE register
	if (psoc_set_offset_register(verbose, tt_mode_register_offset))
		return ERROR_I2C_ACCESS;

	udelay(DELAY_BETWEEN_SET_OFFSET_REG_AND_RW_REG_VALUE);

	//Read value of HOST MODE register
	if (psoc_read_register(verbose, &tt_mode_register_value))
		return ERROR_I2C_ACCESS;

	if (verbose == VERBOSE_DEBUG)
		printk("Read mode, tt mode register value: 0x%x\n", tt_mode_register_value);

	if (verbose == VERBOSE_DEBUG)
		printk("Exit from read mode\n");

	if ( ( tt_mode_register_value & TT_MODE_REGISTER_BOOTLOADER_MODE_MASK ) != 0)
		return ERROR_TS_REBOOTED;

	return 0;
}

static int psoc_get_system_info(int verbose)
{

	int ret;
	int i;
	char* buff_temp;
	unsigned short temp;

	if (verbose == VERBOSE_DEBUG)
		printk("Enter in get sys info\n");

	buff_temp = (char*) kmalloc(len_i2c_frame_sys_info, GFP_KERNEL);
	if (buff_temp == NULL){
		if (verbose)
			printk(KERN_ERR PSOC_DEVNAME ": Error reading system information, can't allocate a buffer of [%d] bytes\n", len_i2c_frame_sys_info);
		return 1;
	}

	ret = psoc_i2c_recv(buff_temp, len_i2c_frame_sys_info);

	if (ret != len_i2c_frame_sys_info){
		if (verbose)
			printk(KERN_ERR PSOC_DEVNAME ": Error reading system information, return value of i2c read: [%d] instead of [%d]\n", ret, len_i2c_frame_sys_info);
		kfree(buff_temp);
		return ERROR_I2C_ACCESS;
	}

	if ((verbose == VERBOSE_DEBUG) || (verbose == VERBOSE_TUNING)){
		printk("\n");
		printk(PSOC_DEVNAME ":Get system info, raw data received:\n");
		printk("%.2x %.2x %.2x ", header_data_tunning[0], header_data_tunning[1], len_i2c_frame_sys_info);
		for (i=0;i<len_i2c_frame_sys_info;i++){
			printk("%.2x ",buff_temp[i]);
		}
		printk("%.2x %.2x\n", footer_data_tunning[0], footer_data_tunning[1]);
		printk("\n");
	}

	memcpy((void*) &i2c_packet_received_sys_info, (void*) buff_temp, len_i2c_frame_sys_info);

	temp = psoc_swap_lsb_msb((char*) &i2c_packet_received_sys_info.bootloader_version);
	i2c_packet_received_sys_info.bootloader_version = temp;
	temp = psoc_swap_lsb_msb((char*) &i2c_packet_received_sys_info.truetouch_standard_product_version);
	i2c_packet_received_sys_info.truetouch_standard_product_version = temp;
	temp = psoc_swap_lsb_msb((char*) &i2c_packet_received_sys_info.applicationID);
	i2c_packet_received_sys_info.applicationID = temp;

	temp = psoc_swap_lsb_msb((char*) &i2c_packet_received_sys_info.resolution_X);
	i2c_packet_received_sys_info.resolution_X = temp;
	temp = psoc_swap_lsb_msb((char*) &i2c_packet_received_sys_info.resolution_Y);
	i2c_packet_received_sys_info.resolution_Y = temp;

	if (verbose == VERBOSE_DEBUG){

		printk(PSOC_DEVNAME ":i2c packet system info received before resolution reversed:\n");
		printk("HST_MODE: 0x%x\n", i2c_packet_received_sys_info.hst_mode);
		printk("BIST_COMM: 0x%x\n", i2c_packet_received_sys_info.bist_comm);
		printk("BIST_STAT: 0x%x\n", i2c_packet_received_sys_info.bist_stat);

		printk("SILICON_ID0: 0x%x\n", i2c_packet_received_sys_info.siliconID0);
		printk("SILICON_ID1: 0x%x\n", i2c_packet_received_sys_info.siliconID1);
		printk("SILICON_ID2: 0x%x\n", i2c_packet_received_sys_info.siliconID2);
		printk("SILICON_ID3: 0x%x\n", i2c_packet_received_sys_info.siliconID3);
		printk("SILICON_ID4: 0x%x\n", i2c_packet_received_sys_info.siliconID4);
		printk("SILICON_ID5: 0x%x\n", i2c_packet_received_sys_info.siliconID5);
		printk("SILICON_ID6: 0x%x\n", i2c_packet_received_sys_info.siliconID6);
		printk("SILICON_ID7: 0x%x\n", i2c_packet_received_sys_info.siliconID7);

		printk("BOOTLOADER_VERSION: 0x%x\n", i2c_packet_received_sys_info.bootloader_version);
		printk("TRUETOUCH_STANDARD_PRODUCT_VERSION: 0x%x\n", i2c_packet_received_sys_info.truetouch_standard_product_version);
		printk("APPLICATION_ID: 0x%x\n", i2c_packet_received_sys_info.applicationID);
		printk("APPLICATION_VERSION_MAJOR: 0x%x\n", i2c_packet_received_sys_info.application_version_major);
		printk("APPLICATION_VERSION_MINOR: 0x%x\n", i2c_packet_received_sys_info.application_version_minor);
		
		printk("RESOLUTION_X: 0x%x\n", i2c_packet_received_sys_info.resolution_X);
		printk("RESOLUTION_Y: 0x%x\n", i2c_packet_received_sys_info.resolution_Y);
		printk("NUMBER_OF_GESTURES: 0x%x\n", i2c_packet_received_sys_info.number_gestures);

		printk("\n");

	}

	printk(PSOC_DEVNAME ":Cypress Capacitive Touch Screen System Information:\n");
                
	printk("Silicon ID0                         : 0x%x\n", i2c_packet_received_sys_info.siliconID0);
	printk("Silicon ID1                         : 0x%x\n", i2c_packet_received_sys_info.siliconID1);

	printk("Bootloader Version                  : 0x%x\n", i2c_packet_received_sys_info.bootloader_version);
	printk("TrueTouch Standard Product Version  : 0x%x\n", i2c_packet_received_sys_info.truetouch_standard_product_version);
	printk("Application ID                      : 0x%x\n", i2c_packet_received_sys_info.applicationID);
	printk("Application Version                 : %x.%.2x\n", i2c_packet_received_sys_info.application_version_major, i2c_packet_received_sys_info.application_version_minor);

	printk("Resolution X                        : 0x%x\n", i2c_packet_received_sys_info.resolution_X);
	printk("Resolution Y                        : 0x%x\n", i2c_packet_received_sys_info.resolution_Y);
	printk("Number of Gestures Supported        : %d\n", i2c_packet_received_sys_info.number_gestures);

	kfree(buff_temp);

	if (verbose == VERBOSE_DEBUG)
		printk("Exit from get sys info\n");

	return 0;

}


static int psoc_get_raw_counts(int verbose)
{

	int ret;
	int i;
	char* buff_temp;
	unsigned short temp;

	buff_temp = (char*) kmalloc(len_i2c_frame_test_mode, GFP_KERNEL);

	if (buff_temp == NULL){
		if (verbose)
			printk(KERN_ERR PSOC_DEVNAME ": Error reading system information, can't allocate a buffer of [%d] bytes\n", len_i2c_frame_test_mode);
		return 1;
	}

	ret = psoc_i2c_recv(buff_temp, len_i2c_frame_test_mode);

	if (ret != len_i2c_frame_test_mode){
		if (verbose)
			printk(KERN_ERR PSOC_DEVNAME ": Error reading raw counts, return value of i2c read: [%d] instead of [%d]\n", ret, len_i2c_frame_test_mode);
		kfree(buff_temp);
		return ERROR_I2C_ACCESS;
	}

	if ((verbose == VERBOSE_DEBUG) || (verbose == VERBOSE_TUNING)){
		printk("\n");
		printk(PSOC_DEVNAME ":Get test data, raw data received:\n");
		printk("%.2x %.2x %.2x ", header_data_tunning[0], header_data_tunning[1], len_i2c_frame_test_mode);
		for (i=0;i<len_i2c_frame_test_mode;i++){
			printk("%.2x ",buff_temp[i]);
		}
		printk("%.2x %.2x\n", footer_data_tunning[0], footer_data_tunning[1]);
		printk("\n");
	}

	kfree(buff_temp);

	return 0;

}

static ssize_t psoc_mode_show(struct device *dev,
                                     struct device_attribute *attr, char *buf)
{
	ssize_t temp;

	if (down_interruptible(&drv_data->mutex))
		return -ERESTARTSYS;

	temp = sprintf(buf, "%d\n", current_mode);

	up(&drv_data->mutex);

	return temp;
}

static ssize_t psoc_mode_store(struct device *dev,
                                     struct device_attribute *attr,
                                     const char *buf, size_t count)
{
	long i;
	
	if (down_interruptible(&drv_data->mutex))
		return -ERESTARTSYS;

	if (strict_strtoul(buf, 10, &i))
		return -EINVAL;

	psoc_switch_to_other_mode(VERBOSE_LEVEL, (int) i);

	up(&drv_data->mutex);

	return count;
}

static DEVICE_ATTR(mode, 0664, psoc_mode_show, psoc_mode_store);

static struct attribute *psoc_attributes[] = {
	&dev_attr_mode.attr,
	NULL,
};

static struct attribute_group psoc_attr_group = {
	.attrs = psoc_attributes,
};



static void psoc_polling_work(struct work_struct *work)
{
	int err;

	if (down_interruptible(&drv_data->mutex))
		return;

	err = psoc_read_mode(VERBOSE_DISABLED);

	if (err == ERROR_TS_REBOOTED )
		psoc_start_ts_app();
	else if (err == ERROR_I2C_ACCESS)
		psoc_restart_ts();

	up(&drv_data->mutex);

}

static void psoc_read_info_work(struct work_struct *work)
{

	if (down_interruptible(&drv_data->mutex))
		return;

	if (current_mode == OPERATING_MODE){
		if ( psoc_set_offset_register(VERBOSE_LEVEL, HST_MODE_REGISTER_OFFSET) == ERROR_I2C_ACCESS )
			psoc_restart_ts();
		udelay(DELAY_BETWEEN_SET_OFFSET_REG_AND_RW_REG_VALUE);
		//Replace verbose level by verbose tuning for final tuning of ttsp
		if ( psoc_get_finger_position(VERBOSE_LEVEL) == ERROR_I2C_ACCESS)
			psoc_restart_ts();
	} else if (current_mode == SYS_INFO_MODE)  {
		if ( psoc_set_offset_register(VERBOSE_LEVEL, HST_MODE_REGISTER_OFFSET) == ERROR_I2C_ACCESS )
			psoc_restart_ts();
		udelay(DELAY_BETWEEN_SET_OFFSET_REG_AND_RW_REG_VALUE);
		if ( psoc_get_system_info(VERBOSE_LEVEL) == ERROR_I2C_ACCESS )
			psoc_restart_ts();
		if ( psoc_toggle_data_read_bit(VERBOSE_LEVEL) == ERROR_I2C_ACCESS)
			psoc_restart_ts();
		if ( psoc_switch_to_other_mode(VERBOSE_LEVEL, OPERATING_MODE) == ERROR_I2C_ACCESS)
			psoc_restart_ts();
	} else if ( (current_mode == RAW_COUNTS_TEST_MODE) || (current_mode == DIFFERENCE_COUNTS_TEST_MODE) \
			|| (current_mode == BASELINE_TEST_MODE) || (current_mode == IDAC_SETTINGS_TEST_MODE) ) {
		if ( psoc_set_offset_register(VERBOSE_LEVEL, HST_MODE_REGISTER_OFFSET) == ERROR_I2C_ACCESS)
			psoc_restart_ts();
		udelay(DELAY_BETWEEN_SET_OFFSET_REG_AND_RW_REG_VALUE);
		if ( psoc_get_raw_counts(VERBOSE_TUNING) == ERROR_I2C_ACCESS)
			psoc_restart_ts();
		if ( psoc_toggle_data_read_bit(VERBOSE_LEVEL) == ERROR_I2C_ACCESS)
			psoc_restart_ts();
	}
	
	up(&drv_data->mutex);

}


// Timer handler
static void psoc_timer(unsigned long data)
{
	schedule_work(&psoc_i2c.polling_work);
	mytimer.expires = jiffies + msecs_to_jiffies(5000);
	add_timer(&mytimer);
}

static irqreturn_t psoc_irq_handler(int irq, void *dev_id)
{
	schedule_work(&psoc_i2c.read_info_work);
	return IRQ_HANDLED;
};

static void psoc_i2c_init(struct i2c_client *client)
{

	BUG_ON(!client);

	//Init psoc_i2c struct
	memset(&psoc_i2c, 0x00, sizeof(psoc_i2c));
	psoc_i2c.client = client;
	memcpy((void*) &pdata, (void*) client->dev.platform_data, sizeof(psoc_pdata_t));
	INIT_WORK(&psoc_i2c.read_info_work, psoc_read_info_work);
	INIT_WORK(&psoc_i2c.polling_work, psoc_polling_work);
	drv_data = kmalloc(sizeof(psoc_drv_data_t), GFP_KERNEL);

	if(NULL == drv_data) {
		printk(KERN_ERR PSOC_DEVNAME ": Can allocate driver data!\n");
		return;
	}

	dev_set_drvdata(&client->dev, drv_data);
	init_MUTEX(&drv_data->mutex);

	//Init Timer
	init_timer(&mytimer);
	mytimer.function=&psoc_timer;
	mytimer.expires = jiffies + msecs_to_jiffies(5000);

}


static int psoc_i2c_probe(struct i2c_client *client, const struct i2c_device_id *id)
{

	int ret;
	int err;

	len_i2c_frame_test_mode = 216;
	len_i2c_frame_sys_info = sizeof(struct i2c_data_received_sys_info);
	len_i2c_frame_finger_position = sizeof(struct i2c_data_received);


	printk(KERN_INFO PSOC_DEVNAME " :probing driver\n");
	drv_data = NULL;

	printk(KERN_INFO PSOC_DEVNAME " :Psoc Capacitive Touch Screen Driver, (C) 2008 TomTom BV\n" );

	BUG_ON(!client);
	BUG_ON(!client->dev.platform_data);
	psoc_i2c_init(client);

	psoc_cap_ts_input_dev = input_allocate_device();

	if (!psoc_cap_ts_input_dev) {
		printk(KERN_ERR PSOC_DEVNAME ": Badf init_alloc_device()\n");
		kfree(drv_data);
		return -10;
	}

	printk(KERN_INFO PSOC_DEVNAME ": Input device allocated\n");

	/* Announce that the ts will generate absolute coordinates */
	set_bit(EV_ABS, psoc_cap_ts_input_dev->evbit);
	printk(KERN_INFO PSOC_DEVNAME ": Event Abs positions declared\n");

	// Declare max and min of X, Y and pressure
	input_set_abs_params(psoc_cap_ts_input_dev, ABS_X, XMIN, XMAX, 0, 0);
	input_set_abs_params(psoc_cap_ts_input_dev, ABS_Y, YMIN, YMAX, 0, 0);
	input_set_abs_params(psoc_cap_ts_input_dev, ABS_PRESSURE, PRESSUREMIN, PRESSUREMAX, 0, 0);

	input_set_abs_params(psoc_cap_ts_input_dev, ABS_FINGERS, FINGERSMIN, FINGERSMAX, 0, 0);
	input_set_abs_params(psoc_cap_ts_input_dev, ABS_FINGER2_X, XMIN, XMAX, 0, 0);
	input_set_abs_params(psoc_cap_ts_input_dev, ABS_FINGER2_Y, YMIN, YMAX, 0, 0);
	input_set_abs_params(psoc_cap_ts_input_dev, ABS_GESTURE, GESTUREMIN, GESTUREMAX, 0, 0);

	psoc_cap_ts_input_dev->name = "PSOC Cypress Capacitive TouchScreen";
	psoc_cap_ts_input_dev->phys = "input(ts)";
	psoc_cap_ts_input_dev->id.vendor = 0xDEAD;
	psoc_cap_ts_input_dev->id.product = 0xBEEF;
	psoc_cap_ts_input_dev->id.version = 0x0101;

	/* Register with the input subsystem */
	ret = input_register_device(psoc_cap_ts_input_dev);

	if (ret) {
		printk(KERN_ERR PSOC_DEVNAME ": Could not register input device(capacitive touchscreen)!\n");
		kfree(drv_data);
		return -EIO;
	}

	printk(KERN_INFO PSOC_DEVNAME ": Input device registered\n");

	if (psoc_restart_ts())
	{
		printk(KERN_ERR PSOC_DEVNAME ": I2C Communication Impossible with the Bootloader\n" );
                printk(KERN_ERR PSOC_DEVNAME ": Most Probably Wrong Factory Data\n" );
		printk(KERN_ERR PSOC_DEVNAME ": Check the Field /features/ts\n" );
		printk(KERN_ERR PSOC_DEVNAME ": Change the Value 300E into 300E_FW_V2\n" );
                printk(KERN_ERR PSOC_DEVNAME ": Or the Other Way Around\n" );
	}

	if (strncmp(recalibrate, "true", 4) == 0) {
		psoc_re_calibrate();
		return 0;
	}

	printk(KERN_INFO PSOC_DEVNAME ": irq number [%d]\n", client->irq);
	set_irq_type(client->irq, IRQ_TYPE_EDGE_RISING);
	if (request_irq(client->irq, &psoc_irq_handler, IRQ_TYPE_EDGE_FALLING, "psoc_1", NULL)) {
		printk(KERN_ERR PSOC_DEVNAME " Could not allocate IRQ (psoc_1)!\n");
		input_unregister_device(psoc_cap_ts_input_dev);
		kfree(drv_data);
		return -EIO;
	}

	err = sysfs_create_group(&psoc_cap_ts_input_dev->dev.kobj, &psoc_attr_group);
	if (err)
		return err;

	add_timer(&mytimer);

	return 0;
}

static int psoc_i2c_remove(struct i2c_client *client)
{

	printk(KERN_INFO PSOC_DEVNAME ": start remove\n");

	del_timer(&mytimer);

	disable_irq(client->irq);

	free_irq(client->irq, NULL);

	sysfs_remove_group(&psoc_cap_ts_input_dev->dev.kobj, &psoc_attr_group);

	input_unregister_device(psoc_cap_ts_input_dev);
	input_free_device(psoc_cap_ts_input_dev);

	kfree(drv_data);

	printk(KERN_INFO PSOC_DEVNAME ": exit remove\n");

	return 0;
}

static int psoc_i2c_suspend(struct i2c_client *client, pm_message_t mesg)
{
	printk(KERN_INFO PSOC_DEVNAME ": start suspending\n");
	return 0;
}

static int psoc_i2c_resume(struct i2c_client *client)
{
	printk(KERN_INFO PSOC_DEVNAME ": start resuming\n");
	return 0;
}

static const struct i2c_device_id psoc_id[] = {
	{ PSOC_DEVNAME, 0 },
	{ }
};
MODULE_DEVICE_TABLE(i2c, psoc_id);

static struct i2c_driver psoc_i2c_driver=
{
	.id	= I2C_DRIVERID_PSOC_CTSIC,
	.probe	= psoc_i2c_probe,
	.remove	= psoc_i2c_remove,
	.suspend= psoc_i2c_suspend,
	.resume	= psoc_i2c_resume,
	.id_table = psoc_id,
	.driver = {
		.name	= PSOC_DEVNAME,
		.owner	= THIS_MODULE,
	},
};

static int __init psoc_init(void)
{
	int err;

	printk(KERN_INFO PSOC_DEVNAME " :start init\n");

	if ((err = i2c_add_driver(&psoc_i2c_driver))) {
		printk(KERN_ERR PSOC_DEVNAME ": Could Not Be Added. Err Code: [%i]\n", err);
	}

	return err;
}

static void __exit psoc_exit(void)
{
	printk(KERN_INFO PSOC_DEVNAME " :start exit\n");
	i2c_del_driver(&psoc_i2c_driver);
}


MODULE_AUTHOR("Vincent Dejouy <vincent.dejouy@tomtom.com>");
MODULE_DESCRIPTION("Driver for I2C connected PSoc Capacitive Touchscreen IC.");
MODULE_LICENSE("GPL");

module_init(psoc_init);
module_exit(psoc_exit);
module_param(recalibrate, charp, S_IRUGO);
