/* drivers/tomtom/vbusmon.c

Copyright (C) 2009 TomTom BV <http://www.tomtom.com/>
Author: Pepijn de Langen <pepijn.delangen@tomtom.com>

This driver implements power detection on the USB vbus pin

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2 as
published by the Free Software Foundation.
*/

#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/workqueue.h>
#include <linux/device.h>
#include <linux/notifier.h>

#include <plat/tt_vbusmon.h>

#define PFX "vbusmon: "
#define DRIVER_DESC_SHORT "TomTom Vbus monitor"
#define DRIVER_DESC_LONG  "TomTom Vbus monitor, (C) 2009 TomTom BV"

static atomic_t vbus_on = ATOMIC_INIT(0);
static int last_vbus = 0;
static BLOCKING_NOTIFIER_HEAD(vbusmon_chain);
static struct work_struct vbus_changed;

/* Define a worker method used to handle changes asynchronously */
static void vbus_changed_work_handler(struct work_struct *work)
{
	int v;
	v=atomic_read(&vbus_on);
	// Invert last recorded vbus - This is still kept because of suspend issues.
	last_vbus^=1;
	VBUSMON_DBGPRINT("%s: notifying listeners about vbus state (%d)\n", __FUNCTION__, last_vbus);
	blocking_notifier_call_chain(&vbusmon_chain,
#ifndef CONFIG_TOMTOM_USBMODE_QUICKCHANGES
		v,
#else
		last_vbus,
#endif
		 NULL);
#ifdef CONFIG_TOMTOM_USBMODE_QUICKCHANGES
	if (last_vbus != v)
	{
		last_vbus^=1;
		// Vbus changed again
		VBUSMON_DBGPRINT("%s: vbus changed between interrupt and work_handler, sending new state (%d)\n",__FUNCTION__,last_vbus);
		blocking_notifier_call_chain(&vbusmon_chain, last_vbus, NULL);
	}
#endif
}

/**
 * This is basically the interrupt handler, but our arch callback supplies the
 * level of the irq in adition to the usual info
 * Return 0 for success
 **/
static int vbus_handler(int pinstate)
{
	VBUSMON_DBGPRINT("%s: vbus is: %d\n",__FUNCTION__,pinstate);
	atomic_set(&vbus_on, pinstate);
	schedule_work(&vbus_changed);
	return 0;
}

static ssize_t vbusmon_sysfs_read_vbus(struct device *dev, struct device_attribute *attr,char *buf)
{
	return sprintf(buf, "%d\n", atomic_read(&vbus_on));
}

static DEVICE_ATTR(vbus, S_IWUGO | S_IRUGO, vbusmon_sysfs_read_vbus,NULL);

int vbusmon_poll(void)
{
	return atomic_read(&vbus_on);
}

/* Notifier chain */
int vbusmon_register_notifier(struct notifier_block *nb)
{
	int ret;
	VBUSMON_TRACE();
	ret = blocking_notifier_chain_register(&vbusmon_chain,nb);
	nb->notifier_call(nb, atomic_read(&vbus_on), NULL);
	return ret;
}

int vbusmon_unregister_notifier(struct notifier_block *nb)
{
	VBUSMON_TRACE();
	return blocking_notifier_chain_unregister(&vbusmon_chain,nb);
}

static int vbusmon_probe(struct platform_device *pdev)
{
	struct vbus_pdata *pdata;
	int ret;

	VBUSMON_TRACE();
	INIT_WORK(&vbus_changed, vbus_changed_work_handler);
	pdata = (struct vbus_pdata *) pdev->dev.platform_data;
	if ((ret=pdata->config_vbusdetect(vbus_handler))) {
		printk(KERN_ERR PFX " unable to configure vbus_handler (%d)\n", ret);
		return ret;
	}
	ret=device_create_file(&pdev->dev, &dev_attr_vbus);
	if (ret) {
		pdata->cleanup_vbusdetect();
		printk(KERN_ERR PFX " unable to create sysfs file (%d)\n", ret);
		return ret;
	}

	atomic_set(&vbus_on, pdata->poll_vbusdetect());/*set initial state*/
	
	return 0;
}

static int vbusmon_remove(struct platform_device *pdev)
{
	struct vbus_pdata *pdata;
	pdata = (struct vbus_pdata *) pdev->dev.platform_data;
	pdata->cleanup_vbusdetect();
	device_remove_file(&pdev->dev, &dev_attr_vbus);
	return 0;
}

#ifdef CONFIG_PM
static int vbusmon_suspend(struct platform_device *pdev, pm_message_t state)
{
	struct vbus_pdata *pdata;
	pdata = (struct vbus_pdata *) pdev->dev.platform_data;
	pdata->do_suspend();
        return 0;
}

static int vbusmon_resume(struct platform_device *pdev)
{
	struct vbus_pdata *pdata;
	pdata = (struct vbus_pdata *) pdev->dev.platform_data;
	pdata->do_resume();
	atomic_set(&vbus_on, pdata->poll_vbusdetect());
	if (last_vbus != atomic_read(&vbus_on))
		schedule_work(&vbus_changed);
        return 0;
}
#else
#define vbusmon_suspend NULL
#define vbusmon_resume  NULL
#endif /* CONFIG_PM */

static struct platform_driver vbusmon_driver = {
	.probe = vbusmon_probe,
	.remove = vbusmon_remove,
	.suspend = vbusmon_suspend,
	.resume = vbusmon_resume,
	.driver = {
		.owner = THIS_MODULE,
		.name = "tomtom-vbus",
	},
};

static int __init vbusmon_driver_init(void)
{
	int ret;
	ret = platform_driver_register(&vbusmon_driver);
	if (ret) {
		printk(KERN_ERR PFX " unable to register driver (%d)\n", ret);
		return ret;
	}
	printk(KERN_INFO DRIVER_DESC_LONG "(" PFX ")\n");
	return 0;
}

static void __exit vbusmon_driver_exit(void)
{
	platform_driver_unregister(&vbusmon_driver);
	return;
}


EXPORT_SYMBOL(vbusmon_poll);
EXPORT_SYMBOL(vbusmon_register_notifier);
EXPORT_SYMBOL(vbusmon_unregister_notifier);

fs_initcall(vbusmon_driver_init);
module_exit(vbusmon_driver_exit);

MODULE_DESCRIPTION(DRIVER_DESC_LONG);
MODULE_AUTHOR("Pepijn de Langen (pepijn.delangen@tomtom.com)");
MODULE_LICENSE("GPL");
